#include <mega32.h>
#include <stdio.h>
#include <spi.h>
#include <delay.h>  

#include "notes.h"

#include "fftlib.c"
#include "sqrt.c"

#define sample2fix(a) (((int)(a))-127)

void initialize(void);
void main(void);
void task_fft(void);
void apply_thresh(int *,char *,int);
int supermax(int*);

#define NBINS 32  //number of bins in the FFT

char sample;//most recent sample
char samplebuf[NBINS];
char sampleidx;//index into samplebuf for next sample
char gotsamples;//sample synchronization flag

//char spibuf[128];

int xr[NBINS],xi[NBINS],xa[NBINS];//FFT in/out buffers
char tb[NBINS];//temporary buffer for holding thresholded FFT results

float scalefactors[127];


//#define DEBUG_SERIAL  //uncomment this to switch to debug serial output

#define NOTEOFFSET -2

void main(void){
 initialize();
 while(1){
   if(gotsamples==1){
 	 task_fft();
 	}
 }

}


char note, octave,lastnote=255,lastoctave=255,currentnote=255,currentvolume=255,currentoctave=255, midikey, lastmidikey,currentkey;
char fingering,lastfingering = 0,indicatorsum,lastplayednote=255,lastplayedoctave=255;
char volume;
float volfloat;

#define BUFFSIZE 128

//variables for ring buffer
unsigned int in, out,mostrecent,forceflush;
unsigned char buffer[BUFFSIZE];
unsigned char txflag, txbuf;

//main task, does everything that isn't in an interrupt
void task_fft(void){

	int i,max,fftsum,samplesum;
  PORTD.2=1;//turn off debug bandwidth indicator led
 
 note = 255;
 octave = 255;
 fingering = PINC&7;
 for(i=0;i<NBINS;i++){
 	xr[i] = (((int)samplebuf[i] - 128)<<1)+39;
 	xi[i] = 0;
 }
 samplesum = 0;
 for(i=0;i<32;i++) samplesum+=(xr[i]<0)?-1*xr[i]:xr[i];

 FFTfix(xr,xi);
 fftsum=0;
 for(i=0;i<16;i++) fftsum+=xa[i];
 for(i=0;i<16;i++) xa[i] = sqrtfix(multfft(xr[i],xr[i])+multfft(xi[i],xi[i]));
 max = supermax(xa);
 if(max <= 10) max = 10;

 //determine current note based on fingering and fft results.  This is very empirical, so you pretty much have to take it for what it is.
 switch(fingering){
 case 0:
  //OPEN
 apply_thresh(xa,tb,multfft(float2fix(0.6f),max));
 if(tb[4]){// printf("C4");
 	note = _C_; octave=4;
 }else if (tb[7]){// printf("C5");
 	note = _C_; octave=5;
 }else if (tb[9]&&(samplesum>800)){// printf("E5");
 	note = _E_; octave = 5;
 }else if (tb[15]&&(samplesum>1000)){// printf("C6");
 	note = _C_; octave=6;
 }else if (tb[11]) {
 	apply_thresh(xa,tb,multfft(float2fix(0.15f),max));
 	if((tb[6]||(samplesum<750))&&(samplesum<1000)){// printf("G4");
      note = _G_; octave = 4;
    }else{// printf("G5");
      note = _G_; octave = 5;
    }
 }
 break;

 case 1:
 //1
 apply_thresh(xa,tb,multfft(float2fix(0.5f),max));
 if(tb[3]){// printf("A#3");
    note = _AS_; octave = 3;}
 else if (tb[13]){
    apply_thresh(xa,tb,multfft(float2fix(0.3f),max));
    if(tb[7]){// printf("A#4");
      note = _AS_; octave = 4;}
    else{// printf("A#5");
      note = _AS_; octave = 5;}
 }
 else if (tb[8]){// printf("D5");
    note = _D_; octave = 5;}
 else if (tb[10]) {
  indicatorsum = 0;
  if(samplesum < 500){
    apply_thresh(xa,tb,multfft(float2fix(0.1f),max));
    if(tb[9]) indicatorsum++;
    if(tb[11]) indicatorsum++;
    if(tb[12]) indicatorsum++;
    if(tb[13]) indicatorsum++;
    apply_thresh(xa,tb,multfft(float2fix(0.2f),max));
    if(!tb[15]) indicatorsum++;
    if(indicatorsum >4){// printf("F5");//150
      note = _F_; octave = 5;}
    else{//                printf("F4");
      note = _F_; octave = 4;}
  } else {
    apply_thresh(xa,tb,multfft(float2fix(0.1f),max));
    if(tb[5]) indicatorsum++;
    if(tb[15]) indicatorsum++;
    apply_thresh(xa,tb,multfft(float2fix(0.25f),max));
    if(tb[11]) indicatorsum++;
    if((indicatorsum>=1) &&(samplesum < 1000)){// printf("F4");//150
      note = _F_; octave = 4;}
    else{//                printf("F5");
      note = _F_; octave = 5;}
  }
 }

  break;

 case 2:
 //2
 apply_thresh(xa,tb,multfft(float2fix(0.5f),max));
 if(tb[4]){// printf("B3");
 	note = _B_; octave = 3;}
 else if (tb[7]){// printf("B4");
 	note = _B_; octave = 4;}
 else if (tb[9]){// printf("D#5");
 	note = _DS_; octave = 5;}
 else if (tb[10]||tb[11]) {
   if(((xa[11]>xa[10])||(samplesum<700))&&(samplesum<1000)){// printf("F#4");
      	note = _FS_; octave = 4;}
 	 else{// printf("F#5");
     	note = _FS_; octave = 5;}

 }
 else if (tb[14]){// printf("B5");
 	note = _B_; octave = 5; }
 break;

  case 3:
  //12
 apply_thresh(xa,tb,multfft(float2fix(0.35/*0.4*/f),max));
 if(tb[3]){// printf("A3");
 	note = _A_; octave = 3;}
 else if (tb[10]){// printf("E4");
 	note = _E_; octave = 4;}
 else if (tb[8]){// printf("C#5");
 	note = _CS_; octave = 5;}
 else if (tb[13]){
    apply_thresh(xa,tb,multfft(float2fix(0.2f),max));
    if(tb[6]){// printf("A4");
    	note = _A_; octave = 4;}
    else{// printf("A5");
    	note = _A_;octave = 5;}
 }
 break;

 case 5:
  //13
 apply_thresh(xa,tb,multfft(float2fix(0.5f),max));
 if(tb[3]){// printf("G3");
 	note = _G_;octave=3;}
 else if (tb[9]||tb[8]){// printf("D4");
 	note = _D_; octave = 4;}
 break;

 case 7:
  //123
 apply_thresh(xa,tb,multfft(float2fix(0.6f),max));
 if(tb[3]){// printf("F#3");
 	note = _FS_;octave =3;}
 else if (tb[8]){// printf("C#4");  //TODO: it goes wtf, m8
 	note = _CS_;octave= 4;}
 break;

  case 6:
  //23
 apply_thresh(xa,tb,multfft(float2fix(0.25f),max));
 if(tb[3]){// printf("G#3");
 	note = _GS_;octave = 3;}
 else if (tb[12]){
    if(tb[6]){// printf("G#4");
    	note = _GS_;octave = 4;}
    else{// printf("G#5");
    	note = _GS_;octave = 5;}
 } else if (tb[9]){// printf("D#4");
 	note = _DS_;octave = 4;}
 break;
 case 4://none here
 break;
 }
//note should now be decided, and we don't need fft results.  start sampling again.
 gotsamples = 0;

 if((note!=lastnote)||(octave!=lastoctave))
   if(samplesum < 100)
     note = 255;  //if volume is "too low", kill the note

 //compute corresponding midikeys for state variables
 if(note!=255){
 	midikey = (octave+1)*12+note;
 } else {
 	midikey = 255;
 }

 if(lastnote!=255){
 	lastmidikey = (lastoctave+1)*12+lastnote;
 } else {
 	lastmidikey = 255;
 }

 if(currentnote!=255){
 	currentkey = (currentoctave+1)*12+currentnote;
 } else {
 	currentkey = 255;
 }

 //scale volume by frequency/note
 volfloat = 6.0f*(scalefactors[midikey]*samplesum);
 if(volfloat>127)
 	volume = 127;
 else
 	volume = (int)volfloat;

 if(currentkey != 255){ //currently playing something over midi
   if((currentkey == lastmidikey)&&(currentkey==midikey)){
     volfloat = 63.0f*((float)volume)/currentvolume;
     if(volfloat>127) volfloat = 127;
     printf("%c%c%c",0xB0,7,/*(int)volfloat*/volume);//volume update, every cycle
   }

   if((currentkey != lastmidikey)&&(currentkey!=midikey)){
     //turn off note

     #ifdef DEBUG_SERIAL
     printf("-%c%c%d\r\n",NOTE_LABELS[currentnote][0],NOTE_LABELS[currentnote][1],currentoctave);
     #else
     printf("%c%c%c",0x80,currentkey-2,127);
     #endif
     lastplayednote = currentnote;
     lastplayedoctave = currentoctave;
     currentnote = 255;
     currentoctave = 255;

     if((midikey == lastmidikey)&&(midikey!=255)){
       //turn on note
       if((note == lastplayednote)&&(octave<lastplayedoctave)){
       } else {
         currentnote = note;
         currentoctave = octave;
         currentvolume = volume;
       #ifdef DEBUG_SERIAL
       printf("+%c%c%d %d %f %d\r\n",NOTE_LABELS[currentnote][0],NOTE_LABELS[currentnote][1],currentoctave,samplesum,volfloat, volume);
       #else

     printf("%c%c%c",0xB0,7,/*63*/volume);
 	     printf("%c%c%c",0x90,midikey+NOTEOFFSET,/*volume*/127);
 	     #endif
       }
     }
   }
 } else { //no current note over midi
    if((midikey == lastmidikey)&&(midikey!=255)){
       if((note == lastplayednote)&&(octave<lastplayedoctave)){
       } else {
      //turn on note
     currentnote = note;
     currentoctave = octave;
     currentvolume = volume;
     #ifdef DEBUG_SERIAL
     printf("+%c%c%d %d %f %d\r\n",NOTE_LABELS[currentnote][0],NOTE_LABELS[currentnote][1],currentoctave,samplesum,volfloat, volume);
     #else
     printf("%c%c%c",0xB0,7,/*63*/volume);
 	   printf("%c%c%c",0x90,midikey+NOTEOFFSET,/*volume*/127);
 	   #endif
     }
    }
 }
 
 //while more left, clear the ring buffer
 while(1){
   #asm("cli");//critical section of ring buffer, must be atomic relative to rxc interrupt, otherwise we have a race condition
    if(out != in) {
      if(forceflush){
        mostrecent = in;
        forceflush = 0;
      }
      if(out != mostrecent){
        txflag = 1;
        txbuf = buffer[out];
        out = out + 1;
        if(out >= BUFFSIZE) out = 0;   //Wrap around check
        forceflush = 0;
      } else {
        forceflush = 1;//no status byte yet, but clear next time around if we haven't received it yet
      }
    }
    #asm("sei");

    //Transmit new data
    if(txflag) {
      printf("%c",txbuf);
      txflag = 0;
    }else {
      break;
    }
 }
 
 lastnote = note;
 lastoctave = octave;
 lastfingering = fingering;

 PORTD.2 = 0;//turn on debug bandwidth indicator led
}

//applies a threshold to a buffer and indicates it in a new buffer
void apply_thresh(int *in,char *out,int thresh){
   int i;
   for(i=0;i<16;i++)
   		if(in[i]>=thresh)
   			out[i] = 1;
   		else
   			out[i] = 0;
}

//fing the maximum value in a buffer
int supermax(int* buf){
	int i,max;
	max = 0;
	for(i=1;i<16;i++)
		if(buf[i]>max)
			max = buf[i];
	return max;
}

//initialize hardware and state
void initialize(void){
    int i;

    DDRB = 0;
    DDRD = 0;

  
    //these clock outputs have been disabled since we are not using the Linear active filter blocks
     /* 
	  //set up timer0 for 100 kHz output
	  TCCR0 = 0b00011001;//CTC, CLK/1,Toggle OC0
	  OCR0 = 79;
    DDRB|=0b00001000;
    */

    /*
    //set up timer2 for 20 kHz output
    TCCR2 = 0b00011010;//CTC, CLK/8,Toggle OC2
    OCR2 = 49;
    DDRD|=0b10000000;
     */

    //set up timer1 for 2.048kHz (1.024 kHz square wave)
    TCCR1A = 0b01000000;
    TCCR1B = 0b00001001;
    OCR1A = 7813;
    //OCR1A = 60000;
    TIMSK |= 1<<4;//enable TIM1_COMPA interrupt
    //DDRD|=0b00110000;


    //set up serial port for MIDI
    //uses the clock freq set in the config dialog box
    #ifdef DEBUG_SERIAL
    #define baud 38400
    #else
    #define baud 31250
    #endif
 	  //#define baud 115200
 	  //#define baud 38400
 	  UBRRL= _MCU_CLOCK_FREQUENCY_ /(baud*16L) - 1L;
 	  UCSRB=0x98; // activate UART, RX interrupt
 	  DDRD.0 = 0; //RX
 	  DDRD.1 = 1; //TX

 	  //set up SPI
 	  SPSR = 0;
 	  //SPCR = 0b01010000;
 	  SPCR = 0b01010011;
 	  DDRB |= 0b10110000;
 	  PORTB.4 = 0;

 	 	//set up potentiometer A/D
    ADMUX = 0b00100000;
    ADCSRA = 0b11001111;
    DDRA  = 0;
    //ADCSR.6=1; //begin a conversion

 	  DDRC = 0;
 	  PORTC = 0;
 	  DDRD |= 1<<2;
 	  
    sampleidx = 0;
    gotsamples = 0;
  	
    //printf("\r\nrunning\r\n\r\n");
    fft_init();//initialize fft constants

    for(i=0;i<128;i++) {//initialize volume/frequency scale factors
    	scalefactors[i] = -0.000892f*(i-53)+0.0337f;
    }

    in = 0;
    out = 0;
    mostrecent = 255;
    forceflush = 0;

    #ifdef DEBUG_SERIAL
    for(i=54;i<=84;i++) {
    	printf("%f ",scalefactors[i]);// = -0.00892f*(i-53)+0.0337f;
    }
    printf("\r\n");
    printf("TrumpetSelect\r\n");
    printf("All notes off\r\n");
    #else
    printf("%c%c",0xC0,/*126*/ /*56*/ 57);//General MIDI Trumpet
    printf("%c%c%c",0xB0,123,0); //Turn off all notes controller
    #endif
    #asm("sei");
}


interrupt [USART_RXC] void rxc(void){

    buffer[in] = UDR;
    if(buffer[in]>>7){//we received a status byte
      mostrecent = in;
      forceflush = 0;//so we no longer need to force a flush next time it hits the critical section
    }
    in = in + 1;
    if(in >= BUFFSIZE) in = 0;     //Wrap around check
    if(in == out) {                     //In-Out collision check
      out = out + 1;
      if(out >= BUFFSIZE) out = 0;
    }
}

//simple, make the ADC sample IFF it's allowed to
interrupt [TIM1_COMPA] i_sample(void){
  if(gotsamples == 0){
  	ADCSR.6=1;
  }
}

//read out the sample
interrupt [ADC_INT] i_adc(void){

  sample = ADCH;
  samplebuf[sampleidx] = sample;
  sampleidx++;
  if(sampleidx >= NBINS){
  	sampleidx = 0;
  	gotsamples = 1;
  }

}
